import numpy as np
from scipy.optimize import minimize
from typing import cast

from admmsolver.objectivefunc import *
from admmsolver.matrix import identity, asmatrixtype, PartialDiagonalMatrix

import pytest

def _randn_cmplx(*shape) -> np.ndarray:
    return np.random.randn(*shape) + 1j* np.random.randn(*shape)

def _to_real_array(x):
    if np.iscomplexobj(x):
        return x.view(np.float64).ravel()
    else:
        return x

def _from_real_array(x):
    x = x.reshape((-1, 2))
    x = x.view(np.complex128)
    return x.reshape(x.shape[0:-1])

def _minimize(f, x0, method="BFGS"):
    x0 = _to_real_array(x0)
    res = minimize(
            lambda x: f(_from_real_array(x)),
            x0,
            method=method,
            options={'maxiter': 100000}
        )
    return _from_real_array(res.x)

def test_least_squares():
    """
    Least-squares fit with a dense coefficient matrix
    """
    np.random.seed(100)
    N1, N2 = 4, 2
    alpha = 2.0
    y = _randn_cmplx(N1)
    A = _randn_cmplx(N1, N2)
    h = _randn_cmplx(N2)
    sqrt_mu = _randn_cmplx(N2, N2)
    mu = asmatrixtype(sqrt_mu.T.conjugate() @ sqrt_mu)
    lstsq = LeastSquares(alpha, A, y)
    x = lstsq.solve(h, mu)

    f_all = lambda x: np.real(alpha * np.linalg.norm(y - A @ x)**2 + \
        h.T.conjugate() @ x + x.T.conjugate() @ h + x.conjugate().T @ (mu @ x))
    x_ref = _minimize(f_all, x, "BFGS")
    np.testing.assert_allclose(x, x_ref, rtol=1e-8)
    np.testing.assert_allclose(f_all(x), f_all(x_ref), rtol=1e-8)


def test_least_squares_partial():
    """
    Least-squares fit with a coefficient matrix of type PartialDiagonalMatrix
    """
    np.random.seed(100)
    N1, N2 = 40, 20
    alpha = 2.0
    rest_dim_size = 20
    y = _randn_cmplx(N1)
    A = PartialDiagonalMatrix(_randn_cmplx(N1//rest_dim_size, N2//rest_dim_size), rest_dims=(rest_dim_size,))
    assert A.shape == (N1, N2)
    h = _randn_cmplx(N2)
    sqrt_mu = _randn_cmplx(N2, N2)
    mu = asmatrixtype(sqrt_mu.T.conjugate() @ sqrt_mu)
    lstsq = LeastSquares(alpha, A, y)
    x = lstsq.solve(h, mu)

    f_all = lambda x: \
        np.real(
            alpha * np.linalg.norm(y - A @ x)**2 + h.T.conjugate() @ x + \
                x.T.conjugate() @ h + x.conjugate().T @ (mu @ x))
    x_ref = _minimize(f_all, x, "BFGS")
    np.testing.assert_allclose(x, x_ref, rtol=1e-4)
    np.testing.assert_allclose(f_all(x), f_all(x_ref), rtol=1e-8)


def test_constrained_least_squares():
    """
    Linearly constrained least squares
    """
    np.random.seed(100)

    N1, N2 = 8, 4
    Nc = 2
    alpha = 2.0
    y = _randn_cmplx(N1)
    A = _randn_cmplx(N1, N2)
    h = _randn_cmplx(N2)
    C = _randn_cmplx(Nc, N2)
    D = _randn_cmplx(Nc)
    sqrt_mu = _randn_cmplx(N2, N2)
    mu = asmatrixtype(sqrt_mu.T.conjugate() @ sqrt_mu)
    lstsq = ConstrainedLeastSquares(alpha, A, y, C, D)

    x = lstsq.solve(h, mu)
    assert np.abs(C@x - D).max() < 1e-10

    #FIXME: how to check x?


def test_L1():
    """
    Minimize alpha * |x|_1 + h^+ x + x^+ h + mu x^+ x
    """
    N = 20
    assert N%2 == 0
    h = 0.5*np.arange(-N//2, N//2)
    mu = identity(N)
    alpha = 1.0

    l1 = L1Regularizer(alpha, N)
    x = l1.solve(h, mu)

    # Naive optimization
    for i in range(N):
        f = lambda x: alpha * np.abs(x) + 2*h[i]*x + mu.diagonals[i] * x**2
        x0 = 0.0
        res = minimize(f, x0, method="BFGS")
        assert np.abs(x[i]-res.x[0]) < 1e-5

def test_non_negative():
    """
    Minimize infty * Theta(-x) + h^+ x + x^+ h + mu x^+ x
    """
    h = np.array([0, -10, 10])
    N = h.size
    mu = identity(N)

    func = NonNegativePenalty(N)
    x = func.solve(h, mu)

    step_f = lambda x: x if x >= 0 else 0

    # Naive optimization
    for i in range(N):
        f = lambda x: 1e+5 * step_f(-x) + 2*h[i]*x + mu.diagonals[i] * x**2
        x0 = 0.0
        res = minimize(f, x0, method="BFGS")
        assert np.abs(x[i]-res.x[0]) < 1e-5


def test_L2():
    """
    Minimize alpha * |A x|_2^2 + h^+ x + x^+ h + mu x^+ x
    """
    N = 10
    M = 5
    sqrt_mu = _randn_cmplx(N, N)
    mu = asmatrixtype(sqrt_mu.T.conjugate() @ sqrt_mu)
    alpha = 2.0
    A = _randn_cmplx(M, N)
    h = _randn_cmplx(N)

    l2 = L2Regularizer(alpha, A)
    x = l2.solve(h, mu)

    # Naive optimization
    f = lambda x: alpha * np.linalg.norm(A @ x)**2 + \
        2*np.real(h.conjugate().T @ x) + np.real(x.conjugate().T @ (mu @ x))
    x_ref = _minimize(f, x, method="BFGS")
    np.testing.assert_allclose(cast(np.ndarray, x), x_ref, atol=np.abs(x_ref).max()*1e-5, rtol=0)


def test_semi_positive_definite_penalty():
    np.random.seed(100)
    K = 20
    N = 10

    h = _randn_cmplx(N**2 * K)

    for mu in [
        asmatrixtype(identity(N**2 * K)),
        PartialDiagonalMatrix(ScaledIdentityMatrix(N**2, 1.0), (K,))]:
        p = SemiPositiveDefinitePenalty((N,N,K), axis=2)
        res = p.solve(h, mu)

        x = res.reshape((N,N,K))
        for k in range(K):
            evals, evecs = np.linalg.eigh(x[:,:,k])
            assert all(evals > -1e-10)




from admmsolver.matrix import *
from admmsolver.matrix import MatrixBase, _vecprod, _pad_by_zero
import numpy as np
from typing import Tuple, cast
import pytest

def _randn_cmplx(*shape) -> np.ndarray:
    return np.random.randn(*shape) + 1j * np.random.randn(*shape)


def test_matmal():
    np.random.seed(100)

    # (12, 12) * (12, 4)
    n1, n2, n3 = 12, 12, 4

    left_mat = []
    left_mat.append(DiagonalMatrix(np.ones(n1)))
    left_mat.append(ScaledIdentityMatrix(n1, 1+1j))
    left_mat.append(PartialDiagonalMatrix(_randn_cmplx(3,3), rest_dims=(4,)))
    left_mat.append(DenseMatrix(_randn_cmplx(n1, n2)))

    right_mat = []
    right_mat.append(DenseMatrix(_randn_cmplx(n2, n3)))
    right_mat.append(ScaledIdentityMatrix((n2, n3), 1+1j))
    right_mat.append(PartialDiagonalMatrix(_randn_cmplx(3,1), rest_dims=(4,)))

    for l in left_mat:
        for r in right_mat:
            print(type(l), type(r))
            lr = l @ r
            assert isinstance(lr, MatrixBase)
            np.testing.assert_allclose(lr.asmatrix(), l.asmatrix() @ r.asmatrix())


def test_matmal2():
    np.random.seed(100)

    # (4, 12) * (12, 12)
    n1, n2, n3 = 4, 12, 12

    left_mat = []
    left_mat.append(DenseMatrix(_randn_cmplx(n1, n2)))
    left_mat.append(PartialDiagonalMatrix(_randn_cmplx(1,3), rest_dims=(4,)))

    right_mat = []
    right_mat.append(DiagonalMatrix(np.ones(n3)))
    right_mat.append(ScaledIdentityMatrix(n3, 1+1j))
    right_mat.append(PartialDiagonalMatrix(_randn_cmplx(3,3), rest_dims=(4,)))
    right_mat.append(DenseMatrix(_randn_cmplx(n2, n3)))

    for l in left_mat:
        for r in right_mat:
            print(type(l), type(r))
            lr = l @ r
            assert isinstance(lr, MatrixBase)
            np.testing.assert_allclose(lr.asmatrix(), l.asmatrix() @ r.asmatrix())


def test_mul_transpose_conj():
    np.random.seed(100)

    # (4, 12)
    n1, n2 = 4, 12

    mat = []
    mat.append(DiagonalMatrix(np.ones(n1)))
    mat.append(DiagonalMatrix(np.ones(n1), shape=(n1, n2)))
    mat.append(ScaledIdentityMatrix(n1, 1+1j))
    mat.append(ScaledIdentityMatrix((n1,n2), 1+1j))
    mat.append(PartialDiagonalMatrix(_randn_cmplx(3,3), rest_dims=(4,)))
    mat.append(DenseMatrix(_randn_cmplx(n1, n2)))

    c = 1 + 0.1j
    for m in mat:
        print(type(m), m.shape)
        cm = c * m
        assert isinstance(cm, MatrixBase)
        np.testing.assert_allclose(cm.asmatrix(), c * m.asmatrix())

        cm = m.T
        assert isinstance(cm, MatrixBase)
        np.testing.assert_allclose(cm.asmatrix(), m.asmatrix().T)

        cc = m.conj()
        assert isinstance(cc, MatrixBase)
        np.testing.assert_allclose(cc.asmatrix(), m.asmatrix().conj())



def test_add():
    np.random.seed(100)

    n = 2

    mat = []
    mat.append(DiagonalMatrix(np.ones(n)))
    mat.append(ScaledIdentityMatrix(n, 1+1j))
    mat.append(PartialDiagonalMatrix(_randn_cmplx(n, n), (1,1)))
    mat.append(DenseMatrix(_randn_cmplx(n, n)))

    for m in mat:
        for m2 in mat:
            print(type(m), type(m2))
            m3 = m + m2
            assert isinstance(m3, MatrixBase)
            np.testing.assert_allclose(m3.asmatrix(), m.asmatrix() + m2.asmatrix())

def test_inv():
    np.random.seed(100)

    n = 4

    mat = []
    mat.append(DiagonalMatrix(np.ones(n)))
    mat.append(ScaledIdentityMatrix(n, 1+1j))
    mat.append(PartialDiagonalMatrix(_randn_cmplx(2, 2), (2,)))
    mat.append(DenseMatrix(_randn_cmplx(n, n)))

    for m in mat:
        inv_m = m.inv()
        assert isinstance(inv_m, MatrixBase)
        np.testing.assert_allclose(inv_m.asmatrix() @ m.asmatrix(), np.identity(n), rtol=0, atol=1e-12)

def test_matvec():
    np.random.seed(100)

    n = 4

    mat = []
    mat.append(DiagonalMatrix(np.ones(n)))
    mat.append(ScaledIdentityMatrix(n, 1+1j))
    mat.append(PartialDiagonalMatrix(_randn_cmplx(2, 2), (2,)))
    mat.append(PartialDiagonalMatrix(ScaledIdentityMatrix(2, 1.0), (2,)))
    mat.append(DenseMatrix(_randn_cmplx(n, n)))

    vec = np.ones(n)

    for m in mat:
        print(type(m))
        mv = m @ vec
        assert isinstance(mv, np.ndarray)
        np.testing.assert_allclose(mv, m.asmatrix()@vec)

@pytest.mark.parametrize("n,m", [(2,4),(4,2)])
def test_matvec_rectangular(n, m):
    np.random.seed(100)

    mat = []
    mat.append(DiagonalMatrix(np.ones(min(n,m)), shape=(n,m)))
    mat.append(ScaledIdentityMatrix((n,m), 1+1j))
    mat.append(PartialDiagonalMatrix(_randn_cmplx(n//2, m//2), (2,)))
    mat.append(PartialDiagonalMatrix(
            DiagonalMatrix(_randn_cmplx(min(n//2, m//2)), (n//2, m//2)),
            (2,))
        )
    mat.append(DenseMatrix(_randn_cmplx(n, m)))

    vec = np.ones(m)

    for m in mat:
        print(type(m))
        mv = m @ vec
        assert isinstance(mv, np.ndarray)
        np.testing.assert_allclose(mv, m.asmatrix()@vec)


@pytest.mark.parametrize("n,m", [(2,4),(4,2)])
def test_batched_matvec(n, m):
    np.random.seed(100)
    nbatch = 3

    mat = []
    mat.append(DiagonalMatrix(np.ones(min(n,m)), shape=(n,m)))
    mat.append(ScaledIdentityMatrix((n,m), 1+1j))
    mat.append(PartialDiagonalMatrix(_randn_cmplx(n//2, m//2), (2,)))
    mat.append(PartialDiagonalMatrix(
            DiagonalMatrix(_randn_cmplx(min(n//2, m//2)), (n//2, m//2)),
            (2,))
        )
    mat.append(DenseMatrix(_randn_cmplx(n, m)))

    vec = _randn_cmplx(m, nbatch)

    for m in mat:
        print(type(m))
        mv = m @ vec
        assert isinstance(mv, np.ndarray)
        np.testing.assert_allclose(mv, m.asmatrix()@vec)


def test_matmal_diagonal():
    np.random.seed(100)
    a = DiagonalMatrix(np.random.randn(2), shape=(4,2))
    b = DiagonalMatrix(np.random.randn(2), shape=(2,4))

    ab = cast(DiagonalMatrix, a @ b)
    ab_ref = np.zeros(4)
    ab_ref[0:2] = a.diagonals * b.diagonals

    np.testing.assert_allclose(ab.diagonals, ab_ref)

def test_vecprod():
    np.testing.assert_allclose(
        _vecprod(np.ones(1), np.ones(2), 3),
        np.array([1,0,0])
    )


def test_pad_by_zero():
    np.testing.assert_allclose(
        _pad_by_zero(np.ones(1), 3),
        np.array([1,0,0])
    )


# Copyright (C) 2021-2022 H. Shinaoka and others
# SPDX-License-Identifier: MIT

import numpy as np
from typing import Union, Tuple, Optional, cast
from abc import abstractmethod

class MatrixBase(object):
    def __init__(self) -> None:
        super().__init__()

        self.shape = (1,1)
        self.ndim = 2

    def is_diagonal(self) -> bool:
        return self.shape[0] == self.shape[1]

    def __neg__(self)->'MatrixBase':
        return -1.0 * self

    @abstractmethod
    def asmatrix(self)->np.ndarray:
        pass

    @abstractmethod
    def __mul__(self, other: Union[float, complex, np.float64, np.complex128])->'MatrixBase':
        return NotImplemented

    __rmul__ = __mul__

    @abstractmethod
    def __matmul__(self, other: Union['MatrixBase', np.ndarray])->Union['MatrixBase',np.ndarray]:
        return NotImplemented

    @abstractmethod
    def __add__(self, other: 'MatrixBase')->'MatrixBase':
        return NotImplemented

    @abstractmethod
    def conjugate(self)->'MatrixBase':
        pass

    @property
    @abstractmethod
    def T(self)->'MatrixBase':
        pass

    conj = conjugate

    def __sub__(self, other)->'MatrixBase':
        return self + (-other)

    @abstractmethod
    def inv(self)->'MatrixBase':
        pass

    @abstractmethod
    def hash(self)->int:
        pass

class DenseMatrix(MatrixBase):
    def __init__(self, matrix: np.ndarray) -> None:
        assert isinstance(matrix, np.ndarray)
        assert matrix.ndim == 2
        self.data = matrix
        self.shape = cast(Tuple[int,int], matrix.shape)
        self.ndim = 2

    def hash(self)->int:
        return matrix_hash(self.asmatrix())

    def asmatrix(self)->np.ndarray:
        return self.data

    def inv(self)->'DenseMatrix':
        return DenseMatrix(np.linalg.inv(self.asmatrix()))

    def __neg__(self)->'DenseMatrix':
        return -1.0 * self

    @property
    def T(self):
        return DenseMatrix(self.data.T)

    def conjugate(self):
        return DenseMatrix(self.data.conjugate())

    conj = conjugate

    def __mul__(self, other)->'DenseMatrix':
        if np.isscalar(other):
            return DenseMatrix(self.data * other)
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __matmul__(self, other)->Union['DenseMatrix', np.ndarray]:
        assert self.shape[1] == other.shape[0]
        assert isinstance(other, MatrixBase) or (isinstance(other, np.ndarray) and other.ndim<=2)
        if isinstance(other, np.ndarray):
            return self.asmatrix() @ other
        else:
            if isinstance(other, ScaledIdentityMatrix):
                return self @ other.to_diagonal_matrix()
            elif isinstance(other, DiagonalMatrix):
                res = np.zeros((self.shape[0], other.shape[1]), dtype=np.complex128) # make real if possible
                min_size = min(*other.shape)
                res[:, 0:min_size] = \
                    self.data[:, 0:min_size] * other.diagonals[None,:]
                return DenseMatrix(res)
            else:
                return DenseMatrix(self.data @ other.asmatrix())

    def __add__(self, other)->'DenseMatrix':
        assert isinstance(other, MatrixBase)
        assert self.shape == other.shape

        return DenseMatrix(self.asmatrix() + other.asmatrix())



class ScaledIdentityMatrix(MatrixBase):
    """
    Scaled Identity matrix with a rectangular shape
    """
    def __init__(
            self,
            shape: Union[int, Tuple[int,int]],
            coeff: Union[complex, float, np.float64, np.complex128]
        )->None:
        assert type(coeff) in [complex, float, np.float64, np.complex128], type(coeff)
        self.shape = (0,0) # type: Tuple[int, int]
        if isinstance(shape, int):
            self.shape = (shape, shape)
        elif isinstance(shape, tuple):
            self.shape = shape
        else:
            raise ValueError("Invalid shape value!")
        self.coeff = coeff # type: Union[complex, float, np.float64, np.complex128]
        self.ndim = 2 # type: int

    def hash(self)->int:
        return matrix_hash(self.coeff)

    def asmatrix(self) -> np.ndarray:
        return self.coeff * np.eye(N=self.shape[0], M=self.shape[1])

    def __neg__(self)->'ScaledIdentityMatrix':
        return -1.0 * self

    def inv(self)->'ScaledIdentityMatrix':
        if not self.is_diagonal():
            raise RuntimeError("A rectangular matrix is not invertible!")
        return ScaledIdentityMatrix(self.shape, 1/self.coeff)

    @property
    def T(self) -> 'ScaledIdentityMatrix':
        return ScaledIdentityMatrix((self.shape[1], self.shape[0]), self.coeff)

    @property
    def diagonals(self)->np.ndarray:
        if not self.is_diagonal():
            raise RuntimeError("Diagonals of a rectangular matrix is ill defined!")
        return np.full(self.shape[0], self.coeff)

    def conjugate(self)->'ScaledIdentityMatrix':
        return ScaledIdentityMatrix(self.shape, np.conjugate(self.coeff))

    conj = conjugate

    def __mul__(self, other)->'ScaledIdentityMatrix':
        if type(other) in [complex, float, np.float64, np.complex128]:
            return ScaledIdentityMatrix(self.shape, self.coeff*other)
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __matmul__(self, other: Union[MatrixBase, np.ndarray])->Union[MatrixBase,np.ndarray]:
        assert self.shape[1] == other.shape[0], f"{self.shape} {other.shape}"
        assert isinstance(other, MatrixBase) or isinstance(other, np.ndarray)
        return self.to_diagonal_matrix() @ other

    def __add__(self, other: MatrixBase)->MatrixBase:
        assert isinstance(other, MatrixBase)
        assert self.shape == other.shape, f"{self.shape} {other.shape}"

        if isinstance(other, ScaledIdentityMatrix):
            return ScaledIdentityMatrix(self.shape, self.coeff + other.coeff)
        elif isinstance(other, DiagonalMatrix):
            return DiagonalMatrix(self.coeff*np.ones(self.shape[0]) + other.diagonals)
        elif isinstance(other, PartialDiagonalMatrix):
            return PartialDiagonalMatrix(
                ScaledIdentityMatrix(other.matrix.shape[0], self.coeff) + other.matrix,
                other.rest_dims)
        else:
            return DenseMatrix(self.asmatrix() + other.asmatrix())

    def to_diagonal_matrix(self):
        """ Convert to a diagonal matrix """
        return DiagonalMatrix(self.coeff*np.ones(min(*self.shape)), self.shape)

class DiagonalMatrix(MatrixBase):
    """
    Diagonal matrix
    """
    def __init__(self, diagonals, shape: Optional[Tuple[int,int]] = None)->None:
        assert diagonals.ndim == 1
        self._diagonals = diagonals
        self.ndim = 2
        if shape is None:
            self.shape = (diagonals.size, diagonals.size)
        else:
            self.shape = shape
        assert min(*self.shape) == diagonals.size, f"{self.shape} {diagonals.size}"

    def hash(self) -> int:
        return matrix_hash(self.diagonals)

    @property
    def diagonals(self) -> np.ndarray:
        return self._diagonals

    def __neg__(self)->'DiagonalMatrix':
        return -1.0 * self

    def inv(self) -> 'DiagonalMatrix':
        if not self.is_diagonal():
            raise RuntimeError("Must be a diagonal matrix!")
        return DiagonalMatrix(1/self.diagonals)

    def asmatrix(self) -> np.ndarray:
        mat = np.zeros(self.shape, dtype=self.diagonals.dtype)
        min_size = min(*self.shape)
        for i in range(min_size):
            mat[i,i] = self.diagonals[i]
        return mat

    @property
    def T(self) -> 'DiagonalMatrix':
        return DiagonalMatrix(self.diagonals, shape=(self.shape[1], self.shape[0]))

    def conjugate(self) -> 'DiagonalMatrix':
        return DiagonalMatrix(self.diagonals.conjugate(), self.shape)

    conj = conjugate

    def __mul__(self, other) -> 'DiagonalMatrix':
        if type(other) in [complex, float, np.float64, np.complex128]:
            return DiagonalMatrix(self._diagonals * other, self.shape)
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __add__(self, other) -> MatrixBase:
        assert isinstance(other, MatrixBase)
        assert self.shape == other.shape

        if isinstance(other, DenseMatrix):
            return DenseMatrix(other.asmatrix() + np.diag(self.diagonals))
        elif isinstance(other, ScaledIdentityMatrix):
            return DiagonalMatrix(self.diagonals + np.full(self.diagonals.size, other.coeff))
        elif isinstance(other, DiagonalMatrix):
            return DiagonalMatrix(self.diagonals + other.diagonals)
        elif isinstance(other, PartialDiagonalMatrix):
            return DenseMatrix(self.asmatrix() + other.asmatrix())
        else:
            return NotImplemented


    def __matmul__(self, other: Union[MatrixBase, np.ndarray]) -> Union[MatrixBase, np.ndarray]:
        """ self @ other """
        assert self.shape[1] == other.shape[0]
        assert isinstance(other, MatrixBase) or isinstance(other, np.ndarray)

        if isinstance(other, np.ndarray):
            dtype = np.dtype(type(self._diagonals[0] * other.ravel()[0]))
            if other.ndim == 1:
                res = np.zeros(self.shape[0], dtype=dtype)
                min_len = min(self._diagonals.size, other.size)
                res[0:min_len] = self._diagonals[0:min_len] * other[0:min_len]
                return res
            else:
                rest_dim = other.shape[1:]
                res = np.zeros((self.shape[0],) + rest_dim, dtype=dtype)
                min_len = min(self._diagonals.size, other.size)
                res[0:min_len, ...] = \
                    np.einsum('d,d...->d...', self._diagonals[0:min_len], other[0:min_len, ...], optimize=True)
                return res
        elif isinstance(other, DenseMatrix):
            return DenseMatrix(self._diagonals[:,None] * other.data)
        elif isinstance(other, DiagonalMatrix):
            min_size = min(self.shape[0], other.shape[1])
            return DiagonalMatrix(
                _vecprod(self._diagonals, other._diagonals, min_size),
                (self.shape[0], other.shape[1])
            )
        elif isinstance(other, PartialDiagonalMatrix):
            return DenseMatrix(self._diagonals[:,None] * other.asmatrix())
        elif isinstance(other, ScaledIdentityMatrix):
            return self @ other.to_diagonal_matrix()
        else:
            return NotImplemented

    def __str__(self) -> str:
        return "DiagonalMatrix: " + self.diagonals.__str__()


class PartialDiagonalMatrix(MatrixBase):
    """
    Matrix that can be composed as
        A otimes I.
    """
    def __init__(self, matrix: Union[np.ndarray, MatrixBase], rest_dims: tuple) -> None:
        assert matrix.ndim == 2
        self.matrix = asmatrixtype(matrix)
        self._matrix_cg = matrix.T.conj()
        self.rest_dims = rest_dims
        self.ndim = 2
        self.shape = (matrix.shape[0]*np.prod(rest_dims), matrix.shape[1]*np.prod(rest_dims))

    def hash(self) -> int:
        return matrix_hash(self.matrix)

    def asmatrix(self) -> np.ndarray:
        return np.einsum(
            'IJ,ij->IiJj',
            self.matrix.asmatrix(),
            np.identity(np.prod(self.rest_dims)),
            optimize=True
        ).reshape(self.shape)

    def __neg__(self)->'PartialDiagonalMatrix':
        return -1.0 * self

    def inv(self) -> 'PartialDiagonalMatrix':
        return PartialDiagonalMatrix(self.matrix.inv(), self.rest_dims)

    @property
    def T(self) -> 'PartialDiagonalMatrix':
        return PartialDiagonalMatrix(self.matrix.T, self.rest_dims)

    def conjugate(self) -> 'PartialDiagonalMatrix':
        return PartialDiagonalMatrix(self.matrix.conjugate(), self.rest_dims)

    conj = conjugate

    def __matmul__(self, other) -> Union[np.ndarray, MatrixBase]:
        """ self @ other """
        assert self.shape[1] == other.shape[0]
        assert isinstance(other, MatrixBase) or isinstance(other, np.ndarray)
        if isinstance(other, np.ndarray):
            return self.matvec(other)
        elif isinstance(other, PartialDiagonalMatrix) and self.rest_dims == other.rest_dims:
            return PartialDiagonalMatrix(self.matrix@other.matrix, self.rest_dims)
        else:
            return DenseMatrix(self.asmatrix() @ other.asmatrix())


    def __mul__(self, other) -> 'PartialDiagonalMatrix':
        if type(other) in [float, complex, np.float64, np.complex128]:
            return PartialDiagonalMatrix(self.matrix * other, self.rest_dims)
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __add__(self, other) -> MatrixBase:
        assert isinstance(other, MatrixBase)
        assert self.shape == other.shape

        if isinstance(other, ScaledIdentityMatrix):
            return PartialDiagonalMatrix(
                    self.matrix + ScaledIdentityMatrix(self.matrix.shape, other.coeff),
                    self.rest_dims
                )
        else:
            return DenseMatrix(self.asmatrix() + other.asmatrix())


    def matvec(self, v: np.ndarray) -> np.ndarray:
        r"""
        (a \otimes I) @ v
        v can be a vector or a tensor.
        In the latter case, the matrix applied to the first axis of v.
        """
        return _matvec_impl(self.matrix, v, self.rest_dims)


def _matvec_impl(
        matrix: Union[MatrixBase, np.ndarray],
        v: np.ndarray,
        rest_dims: tuple
    ) -> np.ndarray:

    res_leading_dim = matrix.shape[0] * np.prod(rest_dims)
    res_shape = (0,) # type: Tuple[int,...]
    if v.ndim == 1:
        res_shape = (res_leading_dim,)
    else:
        res_shape = (res_leading_dim,) + v.shape[1:]

    v = v.reshape(matrix.shape[1], *rest_dims, -1)
    if isinstance(matrix, DiagonalMatrix):
        return cast(np.ndarray, matrix @ v).reshape(res_shape)
    elif isinstance(matrix, DenseMatrix):
        return np.tensordot(
            matrix.asmatrix(),
            v,
            axes=(-1,0)
        ).reshape(res_shape)
    elif isinstance(matrix, ScaledIdentityMatrix):
        return (matrix.coeff * v.ravel()).reshape(res_shape)
    else:
        raise RuntimeError(f"Unsupported type{type(matrix)}!")

def identity(n, dtype=np.float64) -> ScaledIdentityMatrix:
    """ Create an identity matrix """
    n = int(n)
    assert isinstance(n, int), n
    return ScaledIdentityMatrix(n, dtype(1.0))


def matrix_hash(a) -> int:
    """ Compute hash of a matrix a"""
    if isinstance(a, np.ndarray):
        return hash(a.data.tobytes()) # This makes a copy
    elif np.isscalar(a):
        return hash(a)
    else:
        return a.hash()

def asmatrixtype(a) -> MatrixBase:
    assert isinstance(a, MatrixBase) or (isinstance(a, np.ndarray) and a.ndim==2)
    if isinstance(a, np.ndarray):
        return DenseMatrix(a)
    return a


def _vecprod(v1: np.ndarray, v2: np.ndarray, size=Optional[int]):
    """
    Elementwise product of two vectors.
    If v1.size < size and v2.size, the result is padded on the right
    so that the returned array is size.
    """
    assert isinstance(v1, np.ndarray)
    assert isinstance(v2, np.ndarray)
    min_size = min(v1.size, v2.size)
    res = v1[0:min_size] * v2[0:min_size]
    return _pad_by_zero(res, size)

def _pad_by_zero(arr: np.ndarray, size: int):
    assert arr.size <= size
    if arr.size == size:
        return arr
    res = np.zeros(size, dtype=arr.dtype)
    res[0:arr.size] = arr
    return res


from admmsolver.util import *

def test_second_deriv_prj():
    """
    f(x) = x^2
    f''(x) = 2
    """
    xmax = 3
    N = 1000
    x = np.linspace(0, np.sqrt(xmax), N)**2 # Non-uniform mesh

    prj = second_deriv_prj(x)
    ypp = prj @ (x**2)
    ypp_ref = np.full(N-2, 2)

    np.testing.assert_allclose(ypp, ypp_ref)


def test_smooth_regularizer_coeff():
    """
    f(x) = x^2
    f''(x) = 2
    """
    omega_min = 0.0
    omega_max = 3.0
    N = 10000
    omega = np.linspace(np.sqrt(omega_min), np.sqrt(omega_max), N)**2 # Non-uniform mesh

    prj = smooth_regularizer_coeff(omega)

    ypp = 2
    assert np.abs(
           np.linalg.norm(prj @ omega**2)**2 - 
           (omega_max-omega_min) * ypp**2
        ) < 1e-2


# Copyright (C) 2021-2022 H. Shinaoka and others
# SPDX-License-Identifier: MIT

import numpy as np
from itertools import product
from typing import Tuple, Union, Optional, List, Sequence, cast, Callable

from .objectivefunc import ObjectiveFunctionBase
from .matrix import MatrixBase, PartialDiagonalMatrix, asmatrixtype
from .util import norm

class EqualityCondition(object):
    """
    Equality condition:
        E1 @ x_{i1} - E2 @ x_{i2} = 0,
    where i != j.
    """
    def __init__(self,
        i1: int,
        i2: int,
        E1: Union[np.ndarray, MatrixBase],
        E2: Union[np.ndarray, MatrixBase]
        ) -> None:
        assert i1 != i2, "i1 != i2!"
        assert E1.shape[0] == E2.shape[0], "Leading dimensions of E1 and E2 do not match!"
        assert E1.ndim == 2
        assert E2.ndim == 2
        E1 = asmatrixtype(E1)
        E2 = asmatrixtype(E2)
        super().__init__()
        self.i1 = i1
        self.i2 = i2
        self.E1 = E1
        self.E2 = E2

    @property
    def size(self) -> int:
        return self.E1.shape[0]

class Model(object):
    def __init__(self,
        functions: Sequence[ObjectiveFunctionBase],
        equality_conditons: Union[tuple, List[EqualityCondition]] =[]
        ) -> None:
        """
        functions: list of instances of subclasses of ObjectiveFunctionBase
            Define the cost function as the sum of the given functions.

        equality_conditions:
            Equality conditions. Using tuples are deprecated.
        """
        for f in functions:
            isinstance(f, ObjectiveFunctionBase)
        self._functions = functions
        self._num_func = len(functions)
        self._E = np.full((self._num_func, self._num_func), None)
        self._EcE = np.full((self._num_func, self._num_func), None)
        self._EcE2 = np.full((self._num_func, self._num_func), None)

        for ie, e in enumerate(equality_conditons):
            try:
                if isinstance(e, tuple):
                    # For backward compatibility
                    self._add_equality_condition(EqualityCondition(*e))
                else:
                    self._add_equality_condition(e)
            except Exception as e:
                print(f"Error occured when adding {ie}-th equality condition!")
                raise e

        for i in range(self._num_func):
            for k in range(self._num_func):
                if self._E[k, i] is None:
                    continue
                self._EcE[k,i] = self._E[i,k].T.conjugate()@ self._E[k,i]
                self._EcE2[k,i] = self._E[k,i].T.conjugate()@ self._E[k,i]


    @property
    def functions(self) -> Sequence[ObjectiveFunctionBase]:
        return self._functions

    @property
    def num_func(self) -> int:
        return self._num_func

    @property
    def E(self) -> np.ndarray:
        """ E_{ij} """
        return self._E

    @property
    def EcE(self) -> np.ndarray:
        """ E[i,k]^dagger E[k,i] at (k, i) """
        return self._EcE

    @property
    def EcE2(self) -> np.ndarray:
        """ E[k,i]^dagger E[k,i] at (k, i)"""
        return self._EcE2

    def _add_equality_condition(self, e: EqualityCondition) -> None:
        """
        Add an equality condition
        """
        assert isinstance(e, EqualityCondition)
        assert e.E1.shape[1] == self._functions[e.i1].size_x, \
            f"{e.E1.shape} {self._functions[e.i1].size_x}"
        assert e.E2.shape[1] == self._functions[e.i2].size_x, \
            f"{e.E2.shape} {self._functions[e.i2].size_x}"
        if self._E[e.i1, e.i2] is not None:
            raise RuntimeError("Duplicate entries in equality_conditions")
        self._E[e.i2, e.i1] = e.E1
        self._E[e.i1, e.i2] = e.E2


# Backward compatibility
Problem = Model


class SimpleOptimizer(object):
    """
    The simplest ADMM solver
    """
    def __init__(self, model: Model, x0=None, mu=None, max_mu: float =1e+3) -> None:
        """
        model:
           Model to be solved.

        x0: None or list of 1D array
           Initial guesses for variables `x_i`
        
        mu: float
           Penalty term

        max_mu:
           Max value of mu
        """
        assert isinstance(model, Model)
        num_func = model.num_func
        self._h = np.full((num_func, num_func), None)
        self._mu = np.full((num_func, num_func), 0.0)
        self._model = model
        self._max_mu = max_mu

        if x0 is not None:
            for i in range(len(x0)):
                assert model._functions[i].size_x == x0[i].size
            self._x = [x_.copy() for x_ in x0]
        else:
            self._x = [np.zeros(model.functions[k].size_x, dtype=np.complex128)
               for k in range(num_func)]
        
        if mu is None:
            mu = 1.0
        for i, j in product(range(num_func), repeat=2):
            if model.E[i,j] is None or i <= j:
                continue
            self._h[i,j] = np.zeros(model.E[i,j].shape[0], dtype=np.complex128)
            self._mu[i,j] = mu
        
        self._primal_residual = [] # type: List[float]
        self._dual_residual = [] # type: List[float]

    
    @property
    def x(self) -> List[np.ndarray]:
        return self._x
    
    
    def __call__(self, x: List[np.ndarray]) -> float:
        """Evaluate the cost function"""
        return float(np.sum([f(x_) for x_, f in zip(x, self._model.functions)]))
    
    def _hk(self, k: int) -> Optional[np.ndarray]:
        """ Compute `h` for optimizing `x_k` """
        res = []

        E = self._model.E
        EcE = self._model.EcE

        # i < k
        for i in range(k):
            if self._h[k,i] is None:
                continue
            res.append(
                E[i,k].T.conjugate() @ self._h[k,i]
                - self._mu[k,i] * (EcE[k,i] @ self._x[i])
                )
            assert res[-1].size == self._model.functions[k].size_x, \
                f"{res[-1].size} {self._model.functions[k].size_x}"

        # k < i
        for i in range(k+1, self._model.num_func):
            if self._h[i,k] is None:
                continue
            res.append(
                    - E[i,k].T.conjugate() @ self._h[i,k]
                    - self._mu[i,k] * (EcE[k,i] @ self._x[i])
                )
            assert res[-1].size == self._model.functions[k].size_x, \
                f"{res[-1].size} {self._model.functions[k].size_x}"

        if len(res) > 0:
            return _sum(res)
        else:
            return None

    def _mu_k(self, k) -> Optional[MatrixBase]:
        """ Compute `mu` for optimizing `x_k` """

        EcE2 = self._model.EcE2

        res = []
        # i < k
        for i in range(k):
            if self._h[k,i] is None:
                continue
            res.append(self._mu[k,i] * EcE2[i,k])

        # k < i
        for i in range(k+1, self._model.num_func):
            if self._h[i,k] is None:
                continue
            res.append(self._mu[i,k] * EcE2[i,k])

        if len(res) > 0:
            return _sum(res)
        else:
            return None

    def check_convergence(self, rtol) -> bool:
        """ Check convergence """
        converged = True
        for i, j in product(range(self._model.num_func), repeat=2):
            if self._model.E[i,j] is None or i <= j:
                continue
            primal1 = cast(np.ndarray, self._model.E[i,j] @ self._x[j])
            primal2 = cast(np.ndarray, self._model.E[j,i] @ self._x[i])
            d_primal = cast(np.ndarray, primal1 - primal2)
            dual1 = self._mu[i,j] * self._model.E[j,i] @ (self._model.E[i,j] @ self._x[j])
            dual2 = self._mu[i,j] * self._model.E[j,i] @ (self._model.E[i,j] @ self._x_old[j])
            d_dual = cast(np.ndarray, dual1 - dual2)
            converged = converged and \
                norm(d_primal)/max(norm(primal1), norm(primal2)) < rtol
            converged = converged and \
                norm(d_dual)/max(norm(dual1), norm(dual2)) < rtol

        return converged

    def residual(self) -> Tuple[float, float]:
        """ Compute primal residual and dual residual"""
        num_func = self._model.num_func
        primal = 0.0
        dual = 0.0
        for i, j in product(range(num_func), repeat=2):
            if self._model.E[i,j] is None or i <= j:
                continue
            primal += \
                float(
                    np.linalg.norm(
                        self._model.E[i,j]@self._x[j] - self._model.E[j,i]@self._x[i]
                    )
                )
            dual += \
                float(
                    np.linalg.norm(
                        self._mu[i,j] * (
                            self._model.E[j,i] @
                            self._model.E[i,j] @ (self._x[j] - self._x_old[j])
                        )
                    )
                )
        return primal, dual


    def update_mu(self, fact_incr: float = 2.0, th_change: float = 10.0) -> None:
        """ Update mu based on primal residual and dual residual"""
        num_func = self._model.num_func
        for i, j in product(range(num_func), repeat=2):
            if self._model.E[i,j] is None or i <= j:
                continue
            primal = \
                np.linalg.norm(
                    self._model.E[i,j]@self._x[j] - 
                       self._model.E[j,i]@self._x[i]
                )
            dual = \
                np.linalg.norm(
                    self._mu[i,j] * (
                        self._model.E[j,i] @
                            self._model.E[i,j] @ (self._x[j] - self._x_old[j])
                    )
                )
            if primal > th_change * dual:
                self._mu[i,j] *= fact_incr
            if dual > th_change * primal:
                self._mu[i,j] /= fact_incr
            self._mu[i,j] = min(self._mu[i,j], self._max_mu)
        

    def solve(
            self,
            niter:int =10000,
            callback=None,
            interval_update_mu:int =100,
            update_h: bool = True,
            rtol: float = 1e-12
        ) -> None:
        for iter in range(niter):
            self.one_sweep(update_h=update_h)
            primal, dual = self.residual()
            self._primal_residual.append(primal)
            self._dual_residual.append(dual)
            if callback is not None:
                callback()
            if self.check_convergence(rtol):
                return
            if iter % interval_update_mu == 0:
                self.update_mu()

    def one_sweep(self, update_h: bool) -> None:
        """Update all variables in a single sweep"""
        self._x_old = [x_.copy() for x_ in self._x]

        # Optimize x
        for k in range(self._model.num_func):
            self._x[k][:] = self._model.functions[k].solve(
                self._hk(k),
                self._mu_k(k)
            )

        # Optimize dual variables
        if update_h:
            for i in range(self._model.num_func):
                for j in range(i):
                    if self._h[i,j] is not None:
                        self._h[i,j] += self._mu[i,j] * (
                            (self._model.E[j,i]@self._x[i]) -
                            (self._model.E[i,j]@self._x[j])
                        )


def _sum(objs):
    assert isinstance(objs, list)
    res = objs[0]
    if len(objs) > 1:
        for x in objs[1:]:
            res = res + x
    return res



import numpy as np
from typing import Callable, cast

def second_deriv_prj(x):
    """
    x: 1D array of foats in strictly increasing order

    Return a dense matrix that approximates the second derivative of a given y(x) as
    y''(x_i) simeq = sum_j P_{ij} x_j
    for i=1, 2, ..., N-1.
    """
    assert all(x[1:] > x[0:-1]), "x must be in increasing order!"

    prj = np.zeros((x.size-2, x.size), dtype=np.float64)
    for i in range(x.size-2):
        ip = i + 1
        dx_forward = x[ip+1] - x[ip]
        dx_backward = x[ip] - x[ip-1]
        coeff = 2/(dx_forward**2 * dx_backward + dx_backward**2 * dx_forward)
        prj[i,ip-1] = coeff * dx_forward
        prj[i,ip  ] = coeff * (-dx_backward - dx_forward)
        prj[i,ip+1] = coeff * dx_backward
    return prj


def smooth_regularizer_coeff(omega):
    """
    omegqa: 1D array of foats in strictly increasing order

    Return a dense matrix that approximates the integral of the squared second derivative of y(omega) as

    int domega |y''(omega)|^2
    = \sum_{i=1}^{N-2} |sum_j P_{ij} y_j|^2,
    """
    assert all(omega[1:] > omega[0:-1]), "omega must be in increasing order!"

    dx = 0.5*(omega[2:] - omega[0:-2])
    prj_second_deriv = second_deriv_prj(omega)
    return np.sqrt(dx)[:,None] * prj_second_deriv

norm = lambda x: cast(float, np.linalg.norm(cast(np.ndarray, x))) # Callable[np.ndarray, float]


# Copyright (C) 2021-2022 H. Shinaoka and others
# SPDX-License-Identifier: MIT

import numpy as np
from .matrix import DenseMatrix, DiagonalMatrix, PartialDiagonalMatrix, ScaledIdentityMatrix, asmatrixtype, matrix_hash, MatrixBase
from typing import Sequence, Union, Optional, Tuple, cast, Iterable
from scipy.sparse.linalg import lgmres, LinearOperator

add = lambda x, y: x + y
matmul = lambda x, y: x @ y
inv = lambda x: np.linalg.inv(x) if isinstance(x, np.ndarray) else x.inv()

def _assert_optional_types(obj, types):
    flag = False
    if obj is None:
        flag = True
    for t in types:
        flag = flag or isinstance(obj, t)
    assert flag

def _assert_types(obj, types):
    flag = False
    for t in types:
        flag = flag or isinstance(obj, t)
    assert flag


class ObjectiveFunctionBase(object):
    """
    Base class for objective function F(x)
    """
    def __init__(self, size_x: int) -> None:
        super().__init__()
        self._size_x = size_x

    @property
    def size_x(self) -> int:
        return self._size_x

    def __call__(self, x: np.ndarray) -> float:
        """Evaluate F(x)"""
        return NotImplemented

    def solve(self,
        h: Union[np.ndarray,None],
        mu: Optional[MatrixBase]) -> np.ndarray:
        """
        Return argmin_x F(x) + h^+ x + x^+ h + x^+ mu x

        h: 1D array
        mu: a positive definite matrix
        """
        return NotImplemented


class LeastSquares(ObjectiveFunctionBase):
    """
    alpha * ||y - A @ x||_2^2
    """
    def __init__(
            self,
            alpha: float,
            A: Union[np.ndarray, MatrixBase],
            y: np.ndarray
        ) -> None:
        assert A.ndim == 2
        assert y.ndim == 1
        assert A.shape[0] == y.size
        _assert_types(A, [np.ndarray, MatrixBase])
        A = asmatrixtype(A)
        super().__init__(A.shape[1])

        self._alpha = alpha
        self._A = A # type: MatrixBase
        self._y = y # type: np.ndarray
        self._Ac = A.conjugate().T
        self._AcA = self._Ac @ A
        self._Nx = A.shape[1]

        # B = (A^+ A + mu)^{-1}
        self._B_cache = (0, DenseMatrix(np.zeros((1,1)))) # type: Tuple[int, MatrixBase]

    def __call__(self, x: np.ndarray) -> float:
        Ax = cast(np.ndarray, self._A @ x) # type: np.ndarray
        y = cast(np.ndarray, self._y) # type: np.ndarray
        diff = cast(np.ndarray, y - Ax) # type: np.ndarray
        return float(self._alpha * np.linalg.norm(diff)**2)

    def _get_B(self, mu: MatrixBase) -> MatrixBase:
        hash_ = matrix_hash(mu)
        if self._B_cache[0] != hash_:
            self._B_cache = (
                hash_,
                cast(MatrixBase, inv(add(self._alpha * self._AcA, mu)))
            )
        return self._B_cache[1]

    def solve(self, h: Optional[np.ndarray] = None, mu: Optional[MatrixBase] = None) -> np.ndarray:
        _assert_optional_types(h, [np.ndarray])
        _assert_optional_types(mu, [MatrixBase])

        if h is None:
            h = np.zeros(self._Nx)
        if mu is None:
            mu = DiagonalMatrix(np.zeros(self._Nx))
        assert h.shape == (self._Nx,)
        assert mu.shape == (self._Nx, self._Nx)
        vec = self._alpha * (self._Ac @ self._y) - h
        assert isinstance(vec, np.ndarray)
        return cast(np.ndarray, self._get_B(mu) @ vec)


class ConstrainedLeastSquares(LeastSquares):
    r"""
    alpha * ||y - A @ x||_2^2 + \infty * ||C@x - D||_2^2 = 0
    """
    def __init__(self,
            alpha: float,
            A: Union[np.ndarray, MatrixBase],
            y: np.ndarray,
            C: Union[np.ndarray, MatrixBase],
            D: np.ndarray
        ) -> None:
        assert A.ndim == 2
        assert y.ndim == 1
        assert C.ndim == 2
        assert D.ndim == 1
        assert A.shape[0] == y.size
        assert A.shape[1] == C.shape[1]
        assert C.shape[0] == D.size
        _assert_types(A, [np.ndarray, MatrixBase])
        _assert_types(C, [np.ndarray, MatrixBase])
        super().__init__(alpha, asmatrixtype(A), y)

        self._C = DenseMatrix(C) if isinstance(C, np.ndarray) else C # type: MatrixBase
        self._D = D

    def solve(self, h: Optional[np.ndarray] = None, mu: Optional[MatrixBase] = None) -> np.ndarray:
        _assert_optional_types(h, [np.ndarray])
        _assert_optional_types(mu, [MatrixBase])

        if h is None:
            h = np.zeros(self._Nx)
        assert h.shape == (self._Nx,)
        if mu is None:
            mu = ScaledIdentityMatrix(self._Nx, 0.0)
        assert mu.shape == (self._Nx, self._Nx)
        Ch = self._C.conjugate().T # type: MatrixBase
        B =  self._get_B(mu)
        xi1 = cast(np.ndarray, B @ (self._alpha * (self._Ac @ self._y) - h))
        assert isinstance(xi1, np.ndarray)
        xi2 = - (B @ Ch)
        tmp1 = inv(self._C @ xi2) # Union[MatrixBase, np.ndarray]
        tmp2 = cast(np.ndarray, self._D - self._C @ xi1)
        assert isinstance(tmp2, np.ndarray)
        nu = tmp1 @ tmp2
        return xi1 + xi2 @ nu


class L1Regularizer(ObjectiveFunctionBase):
    """
    L1 regularization
        F(x) = alpha * |x|_1
    """
    def __init__(self, alpha: float, size_x: int) -> None:
        assert isinstance(size_x, int), type(size_x)
        super().__init__(size_x)
        assert alpha > 0
        self._alpha = alpha

    def __call__(self, x) -> float:
        return self._alpha * np.sum(np.abs(x))

    def solve(
        self,
        h: Optional[np.ndarray] = None,
        mu: Optional[MatrixBase] = None) -> np.ndarray:
        """
        x <- argmin_x alpha * |x|_1 + h^+ x + x^+ h + mu x^+ x

        This function works only if all the following conditions are met:
          * h and x are real vectors
          * mu is a diagonal matrix
        Return a real vector.
        """
        _assert_types(h, [np.ndarray])
        assert isinstance(mu, DiagonalMatrix) or isinstance(mu, ScaledIdentityMatrix)

        if h is None:
            raise ValueError("h must not be None!")
        if mu is None:
            raise ValueError("mu must not be None!")
        if np.iscomplexobj(h):
            h = h.real
        return _softmax(-(h/mu.diagonals), 0.5*self._alpha/mu.diagonals)


class L2Regularizer(ObjectiveFunctionBase):
    """
    L2 regularization
        F(x) = alpha * |A x|_2^2
    """
    def __init__(self, alpha: float, A: Union[np.ndarray, MatrixBase]):
        _assert_optional_types(A, [np.ndarray, MatrixBase])
        A = asmatrixtype(A)
        super().__init__(A.shape[1])
        assert alpha > 0
        self._alpha = alpha
        self._A = A
        self._AcA = matmul(A.conjugate().T, A)

        # B = (A^+ A + mu)^{-1}
        self._B_cache = (0, DenseMatrix(np.zeros((1,1))))

    def __call__(self, x: np.ndarray):
        return self._alpha * np.linalg.norm(matmul(self._A, x))**2

    def _get_B(self, mu: MatrixBase):
        hash_ = matrix_hash(mu)
        if self._B_cache[0] != hash_:
            self._B_cache = (
                hash_,
                inv(add(self._alpha * self._AcA, mu))
            )
        return self._B_cache[1]

    def solve(self,
        h: Optional[np.ndarray] = None,
        mu: Optional[MatrixBase] = None):
        """
        x <- argmin_x alpha * x^+ (A^+ A) x + h^+ x + x^+ h + x^+ mu x
            = - (alpha A^+ A + mu)^{-1} h

        """
        _assert_optional_types(h, [np.ndarray])
        _assert_optional_types(mu, [MatrixBase])
        if mu is None:
            mu = ScaledIdentityMatrix(self._A.shape[1], 0.0)
        if h is None:
            return np.zeros(self._A.shape[1])
        else:
            return - (self._get_B(mu) @ h)


class NonNegativePenalty(ObjectiveFunctionBase):
    """
    Non-negative penalty term
        F(x) = infty * Theta(-x)
    """
    def __init__(self, size_x: int):
        super().__init__(size_x)

    def __call__(self, x: np.ndarray):
        return 0.0

    def solve(self, h: Optional[np.ndarray] = None, mu: Optional[MatrixBase] = None):
        """
        This function works only if all the following conditions are met:
          * h and x are real vectors
          * mu is a diagonal matrix
        Return a real vector.
        """
        assert isinstance(h, np.ndarray)
        assert isinstance(mu, DiagonalMatrix) or isinstance(mu, ScaledIdentityMatrix)
        if h is None:
            h = np.zeros(self.size_x)
        elif np.iscomplexobj(h):
            h = h.real
        if mu is None:
            raise ValueError("mu must not be None!")
        return _project_plus(-(h/mu.diagonals))


class SemiPositiveDefinitePenalty(ObjectiveFunctionBase):
    """
    Penalty term for negative eigenvalues of x

    1) Reshape x into a three-way tensor
    2) Along a given axis, we compute eigenvalues and remove negative ones (we assume hermition matrices).
    """
    def __init__(self, shape: Union[Sequence,np.ndarray], axis: int):
        assert len(shape) == 3
        super().__init__(np.prod(shape))
        self._shape = shape
        self._axis = axis

    def __call__(self, x: np.ndarray):
        return 0.0

    def solve(self, h : Optional[np.ndarray] = None, mu: Optional[MatrixBase] = None):
        """
        This function works only if mu is a diagonal matrix
        mu must be regarded as a diagonal matrix
        """
        assert isinstance(h, np.ndarray)
        assert isinstance(mu, DiagonalMatrix) or isinstance(mu, ScaledIdentityMatrix) or \
            (isinstance(mu, PartialDiagonalMatrix) and isinstance(mu.matrix, ScaledIdentityMatrix)) or \
            (isinstance(mu, PartialDiagonalMatrix) and isinstance(mu.matrix, DiagonalMatrix))
        if mu is None:
            raise ValueError("mu must not be None!")

        diagonals = None
        if isinstance(mu, DiagonalMatrix) or isinstance(mu, ScaledIdentityMatrix):
            diagonals = mu.diagonals
        elif isinstance(mu.matrix, ScaledIdentityMatrix):
            diagonals = np.full(mu.shape[0], mu.matrix.coeff)
        elif isinstance(mu.matrix, DiagonalMatrix):
            diagonals = np.einsum('i,j->ij', mu.matrix.diagonals, np.ones(np.prod(mu.rest_dims))).ravel()
        assert diagonals is not None
        assert diagonals.ndim == 1, diagonals.shape


        if h is None:
            h = np.zeros(np.prod(self._shape))
        elif np.iscomplexobj(h):
            h = h.real

        assert diagonals.size == h.size

        x_ = (-(h/diagonals)).reshape(self._shape)
        x_ = np.moveaxis(x_, self._axis, 0)
        for i in range(x_.shape[0]):
            evals, evecs = np.linalg.eigh(x_[i,:,:])
            idx = evals >= 0
            U = evecs[:, idx]
            x_[i,:,:] = U @ (evals[idx,None] * U.conjugate().T)
        return np.moveaxis(x_, 0, self._axis).ravel()


def _project_plus(x):
    res = x.copy()
    res[x < 0] = 0
    return res

def _softmax(y: np.ndarray, lambda_: np.ndarray):
    """
    Softmax function

    y - lambda_ (y >  lambda_)
    y + lambda_ (y < -lambda_)
    0           (otherwise)
    """
    assert isinstance(y, np.ndarray) and y.ndim == 1
    assert isinstance(lambda_, np.ndarray) and lambda_.ndim == 1
    assert (np.asarray(lambda_) > 0).all()
    assert y.size == lambda_.size
    res = np.zeros(y.size)

    idx = y > lambda_
    res[idx] = y[idx] - lambda_[idx]

    idx = y < -lambda_
    res[idx] = y[idx] + lambda_[idx]

    return res



import numpy as np
from scipy.optimize import minimize
from scipy.optimize._trustregion_constr.equality_constrained_sqp import equality_constrained_sqp

from admmsolver.objectivefunc import LeastSquares, L1Regularizer, L2Regularizer
from admmsolver.optimizer import SimpleOptimizer, Model
from admmsolver.matrix import identity, DiagonalMatrix


def _randn_cmplx(*shape):
    return np.random.randn(*shape) + 1j* np.random.randn(*shape)

def test_LASSO():
    """
    |y - A * x|^2 + alpha |x|_1,
    where
       y = [2.]
       A = [[2., 1.]]
       alpha = 0.1
      
    The solution is x simeq (1, 0).

    This problem can be solved by minimizing
       |y - A * x0|^2 + alpha |x1|_1
    subject to x0 - x1 = 0.
    """
    y = np.array([2])
    A = np.array([[2, 1]])
    alpha = 0.1

    # Naive optimization
    f = lambda x: np.linalg.norm(y - A @ x)**2 + alpha * np.sum(np.abs(x))
    res = minimize(f, x0=np.array([1.1,0]), method="Nelder-Mead", options={"xatol": 1e-10})
    assert res.success 
    x_ref = res.x

    # ADMM
    lstsq = LeastSquares(1.0, A, y)
    l1 = L1Regularizer(alpha, A.shape[1])
    equality_conditions = [
       (1, 0, identity(2), identity(2))
    ]
    p = Model([lstsq, l1], equality_conditions)
    opt = SimpleOptimizer(p)

    assert np.abs(opt(2*[x_ref]) - f(x_ref)) < 1e-10
    opt.solve(100)
    x_res = opt.x
    for x in x_res:
        np.testing.assert_allclose(x, x_ref, atol=1e-10)

def test_basis_pursuit():
    # Fig. 3 of J. Otsuki et al., JPSJ89, 012001 (2020)
    # https://github.com/SpM-lab/CS-tools/tree/master/jpsj-review/basis_pursuit
    # Dimension of the signal
    N = 1000

    M = 100
    K = 20    
    seed = 1234
    np.random.seed(seed)
    A = np.random.randn(M,N)
   
    #Make answer vector
    xanswer = np.zeros(N)
    xanswer[:K] = np.random.randn(K)
    xanswer = np.random.permutation(xanswer)
   
    y_calc = np.dot(A, xanswer)    

    lstsq = LeastSquares(1.0, A, y_calc)
    l1 = L1Regularizer(1e-1, A.shape[1])
    equality_conditions = [
      (1, 0, identity(N), identity(N))
    ]
    p = Model([lstsq, l1], equality_conditions)
    opt = SimpleOptimizer(p)

    niter = 100
    opt.solve(niter)

    np.testing.assert_allclose(opt.x[0], xanswer, atol=1e-2*np.abs(xanswer).max(), rtol=0)


def test_ridge():
    """
    |y - A * x|^2 + alpha |B x|_2,
    The solution is
       - (A^+ A + B^+B)^{-1} A^+ y
    """
    N1, N2, N3 = 2, 2, 1
    np.random.seed(100)
    y = _randn_cmplx(N1)
    A = _randn_cmplx(N1, N2)
    B = _randn_cmplx(N3, N2)
    alpha = 1

    # ADMM
    lstsq = LeastSquares(1.0, A, y)
    l2 = L2Regularizer(alpha, B)
    equality_conditions = [
       (1, 0, identity(N2), identity(N2))
    ]
    model = Model([lstsq, l2], equality_conditions)
    opt = SimpleOptimizer(model)
    opt.solve(niter=100, update_h=True)

    x_ref = np.linalg.inv(A.conjugate().T @ A + alpha * B.conjugate().T @ B) @ A.conjugate().T @ y
    np.testing.assert_allclose(opt.x[0], x_ref, atol=np.abs(x_ref).max()*1e-8)